
#include "GenotypeInferrer.h"


GenotypeInferrer::GenotypeInferrer(Pedigree &_ped, IBDTable &_ibd)
	: ped(_ped), ibd(_ibd)
{
	// ped = _ped; ibd = _ibd;
}

/** Infers some of the children's genotypes.
	These will be filled into the pedigree. */
void GenotypeInferrer::inferGenotypes() {
	for(map<pair<string, string>, Person *>::iterator i =
			ped.peopleByID.begin();
			i != ped.peopleByID.end(); i++)
		inferGenotypes(i->second);
}

/** Infers genotypes for one person.
	This will set the parent and grandparent variables,
and return without doing anything if the parents and
grandparents aren't available.
  If parents and grandparents are available, loops
through the markers, replacing genotypes if possible. */
void GenotypeInferrer::inferGenotypes(Person *_p) {
	p = _p;
	
	// Sets the parent variables.  If any of these are missing,
	// return without inferring anything.
	// XXX slightly funny-looking; relies on the sequencing
	// behaviour of &&.  Should return if anything's NULL.
	if (!( (father = p->father) && (mother = p->mother) &&
			(pgf = father->father) && (pgm = father->mother) &&
			(mgf = mother->father) && (mgm = mother->mother) ))
		return;

	// Loop through the markers.  Admittedly, without type inference,
	// this is quite ugly.  (Possibly I should loop by marker name,
	// but that would jump around in memory a lot, assuming the pedigree
	// has markers approximately sorted by chromosome & location.)
	for(map<string, map<int, Marker *> >::iterator chrom
			= ped.markersByLoc.begin(); chrom != ped.markersByLoc.end(); chrom++)
		for(map<int, Marker *>::iterator loc = chrom->second.begin();
				loc != chrom->second.end(); loc++)
			inferGenotype(*(loc->second));
}

/** Infers a genotype for one person at one marker. */
void GenotypeInferrer::inferGenotype(Marker &m) {
	int i = m.genotypeIndex;
	
	// if the genotype is already known, do nothing
	if (! p->genotype[i].isMissing() )
		return;

	short paternalAllele = inferAllele(m, pgf, pgm, father);
	short maternalAllele = inferAllele(m, mgf, mgm, mother);
	
	// if we could determine both halves of the genotype, fill it in
	// nope, now, if one allele can be inferred, we do so, even if
	// the other one is missing.
	// if (paternalAllele != 0 && maternalAllele != 0)
  p->genotype[i] = Genotype(paternalAllele, maternalAllele);
}

/** Infers one allele. */
short
GenotypeInferrer::inferAllele(Marker &m,
Person *grandfather, Person *grandmother, Person *parent) {
	int i = m.genotypeIndex;
	
	// If the parent genotype is missing, we can only infer something
	// if the grandparents are all homozygous (and the same).  Perhaps
	// we should include this case, but for now, we skip it.
	if (parent->genotype[i].isMissing())
		return 0;
		
	// If the parent is homozygous, we're in luck...
	if (parent->genotype[i].isHomozygous())
		return parent->genotype[i].a;
		
	// at this point, we need phase information.
	bool grandfatherSharing =
		ibd.getIBD(p, grandfather, m.chrom, m.loc).p1() == 1.0;
	bool grandmotherSharing =
		ibd.getIBD(p, grandmother, m.chrom, m.loc).p1() == 1.0;
	// only proceed if exactly one of these is true
	if (! (grandfatherSharing ^ grandmotherSharing))
		return 0;

	// standardize which grandparent is which
	Person *ibdGrandparent = grandfatherSharing ? grandfather : grandmother;
	Person *nonIbdGrandparent = grandmotherSharing ? grandfather : grandmother;

	// if the IBD grandparent is homozygous, we're done
	if ( ! ibdGrandparent->genotype[i].isMissing() &&
			ibdGrandparent->genotype[i].isHomozygous() )
		return ibdGrandparent->genotype[i].a;

  // At this point, we know the parent, and the IBD grandparent,
  // both are heterozygous.  If the non-IBD grandparent is
  // homozygous, then we know that the other allele from the
  // IBD grandparent is what was transmitted.
 	if ( ! nonIbdGrandparent->genotype[i].isMissing() &&
			nonIbdGrandparent->genotype[i].isHomozygous() ) {
			
		// compute the allele "other than" what the allele
		// from the non-IBD grandparent was.  (The non-IBD grandparent
		// was homozygous, so it doesn't matter which allele we look at.)
		if (nonIbdGrandparent->genotype[i].a ==
				ibdGrandparent->genotype[i].a)
			return ibdGrandparent->genotype[i].b;
		else
			return ibdGrandparent->genotype[i].a;
	}

	// at this point, we can't infer the genotype
	return 0;
}
