#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <algorithm>
#include <iostream>
#include <vector>

#include "IBDTable.h"
#include "Pedigree.h"
#include "Person.h"
#include "Utils.h"

using namespace std;

/** Reads in IBD information. */
IBDTable::IBDTable(istream &ibdFile, Pedigree &p) {
  string family, ID1, ID2, markerName;
  double p0, p1, p2;

	// read first line, and check that it makes sense
	if (!readFirstLine(ibdFile)) {
		printf("error: unknown IBD file format\n");
		exit(1);
	}

  while (isalnum(ibdFile.peek())) {
    ibdFile >> family >> ID1 >> ID2 >> markerName;
    
		// standardize marker name to lower case, and look up marker
		std::transform(markerName.begin(), markerName.end(),
			markerName.begin(), ::tolower);
    Marker *m = p.markersByID[markerName];

		IBD i = readIBD(ibdFile, m);

		// find the two people in the Pedigree
		Person *person1 = p.peopleByID[ pair<string,string>(family, ID1) ];
		Person *person2 = p.peopleByID[ pair<string,string>(family, ID2) ];

if (!person1) {
  cerr << "no IBD info for person with family=" << family << " ID=" << ID1 << endl;
  exit(1);
}
if (!person2) {
  cerr << "no IBD info for person with family=" << family << " ID=" << ID2 << endl;
  exit(1);
}

		// note that the people are "sorted" in the pair, so that
		// we only need make one lookup
		pair<Person *, Person *> a = sortPeople(person1, person2);

		// We only store IBD information if it's definite about sharing.
		if (_isExtendedFile) {
			// if this is an extended IBD file, only store IBD if it
			// definitely indicates sharing (or not) on either side.
			if (i.paternal() == 0 || i.paternal() == 1 ||
					i.maternal() == 0 || i.maternal() == 1)
				ibd[a][m->chrom][m->loc] = i;
		}
		else
			// if it's not extended IBD, only store it if
			// sharing is exactly 0, 1, or 2.
			if (i.p0() == 1 || i.p1() == 1 || i.p2() == 1)
				ibd[a][m->chrom][m->loc] = i;
	}
}

/** Gets the estimated IBD for two people, at some location. */
IBD IBDTable::getIBD(Person *a, Person *b, string chrom, int loc) {
	// this will be returned if we don't know the IBD
	IBD unknownIBD(NULL, 0, 0, 0, 0);
	
	// XXX ugly hack for testing!
//	return unknownIBD;
	
	// FIXME: special cases like a == b, a == parent of b (or vice versa),
	// etc., shouldn't need a table lookup
	
//	cout << "size of ibd = " << ibd.size() << endl;
//	cout << "size of ibd[...] = " << ibd[sortPeople(a,b)].size() << endl;
//	cout << "size of ibd[...][...] = " << ibd[sortPeople(a,b)][chrom].size() << endl;

	
	// First, get the map of IBD for those people and that chromosome
	map<int, IBD> i = ibd[sortPeople(a,b)][chrom];
	map<int, IBD>::iterator i1 = i.lower_bound(loc);

//	cout << "getting IBD for " << a->family << " " << a->id << " and " <<
//		b->family << " " << b->id << " at " << chrom << ":" << loc << endl;
//	cout << "lower bound is at " << i1->first << endl;
	
	// Did we find that location exactly?
	if (i1->first == loc)
		return i1->second;
		
	// Are we outside the markers?
	if (i1 == i.begin() || i1 == i.end()) {
//		cout << "off ends" << endl;
		return unknownIBD;
	}

	// At this point, we should be between markers with known IBD, which
	// should be between i1's predecessor (i0, which should be defined)
	// and i1.
	map<int, IBD>::iterator i0 = i1;
	i0--;

//	cout << "we're between " << i0->first << " and " << i1->first << endl;
	
	// For now, we assume that we only know IBD if there were informative
	// markers on both sides.  (We could assume that no crossovers happened
	// outside of end markers.)
	// Also, for now, we don't interpolate IBD when it changes.
	if (i0->second == i1->second)
		return i0->second;
	else
		// FIXME: should depend on the relationship between a and b
		// for now, we return something which should indicate "not known"
		return unknownIBD;		
}

/** Gets the sharing on one side (paternal or maternal), at
	some location.  paternalSide determines which side.
	This function is a temporary hack;
	at some point we'll probably switch to just using multipoint IBD.
	But since, for now, we're using "cases where IBD is definitely known",
	this is needed for the sib-inference case -- if the maternal side is
	known at several adjacent markers, we don't care if the paternal side
	is unknown. */
double IBDTable::getSharingOneSide
(Person *a, Person *b, string chrom, int loc, bool paternalSide) {
	
	// First, get the map of IBD for those people and that chromosome
	map<int, IBD> i = ibd[sortPeople(a,b)][chrom];
	map<int, IBD>::iterator i1 = i.lower_bound(loc);

	// Did we find that location exactly?
	if (i1->first == loc)
		return paternalSide ? i1->second.paternal() : i1->second.maternal();
		
	// Are we outside the markers?
	// For now, we assume that we only know IBD if there were informative
	// markers on both sides.  (We could assume that no crossovers happened
	// outside of end markers.)
	if (i1 == i.begin() || i1 == i.end()) {
		return 0.5;     // "unknown"
	}

	// find sharing to the left of the marker by scanning backwards
	map<int, IBD>::iterator i0 = i1;
	i0--;
	// At this point, the marker should be between i0 and i1, and
	// both i0 and i1 should have valid IBD entries (although
	// possibly not informative IBD.)
	double shareLeft =
		paternalSide ? i0->second.paternal() : i0->second.maternal();
	while (!(shareLeft == 0 || shareLeft == 1)
			&& i0 != i.begin() && i0 != i.end()) {
		shareLeft = paternalSide ? i0->second.paternal() : i0->second.maternal();
		i0--;
	}
	
	// and similarly, scan forward to find sharing to the right of the marker
	double shareRight =
		paternalSide ? i1->second.paternal() : i1->second.maternal();
	while (!(shareRight == 0 || shareRight == 1)
			&& i1 != i.begin() && i1 != i.end()) {
		shareRight = paternalSide ? i1->second.paternal() : i1->second.maternal();
		i1++;
	}
	
///	cout << "we're between " << i0->first << " and " << i1->first << endl;
	
	// For now, we assume that we only know IBD if the nearest
	// informative IBD on each side is the same.
	if (shareLeft == shareRight)
		return shareLeft;
	else
		// FIXME: should depend on the distance between a and b
		// for now, we return something which should indicate "not known"
		return 0.5;
}

/** Standardizes a pair of two people so that they're in sorted order,
	first by family, then individual. */
pair<Person *, Person *> IBDTable::sortPeople(Person *a, Person *b) {
	if (a->family < b->family)
		return pair<Person *, Person *>(a, b);
	if (a->family > b->family)
		return pair<Person *, Person *>(b, a);
		
	// at this point, they must be in the same family
	assert(a->family == b->family);   // or, they had better be
	if (a->id < b->id)
		return pair<Person *, Person *>(a, b);
	else
		return pair<Person *, Person *>(b, a);
}

double IBDTable::probOfCrossover(double cmDistance) {
	return 0.5 * (1.0 - exp(-2.0 * 100.0 * cmDistance));
}
	
/** This is a pretty superficial test.  For now, we just check that
	* whichever line is what MERLIN would have outputted.  If it's
	* not, we print a warning, but continue. */
bool IBDTable::readFirstLine(istream &is) {
	// XXX seems like there should be an easier way to do this
	char s[255];
	is.getline(s, 255);
	string firstLine(s);

	if (firstLine ==
"FAMILY ID1 ID2 POSITION S01 S02 S03 S04 S05 S06 S07 S08 S09 S10 S11 S12 S13 S14 S15") {
		_isExtendedFile = true;
		return true;
	}
	
	if (firstLine ==
"FAMILY ID1 ID2 MARKER P0 P1 P2") {
		_isExtendedFile = false;
		return true;
	}
	
	fprintf(stderr, "warning: IBD file may not be in correct format (first line looks wrong.)\n");
	_isExtendedFile = false;
	return true;   // XXX arguably, this should be more particular
}

/** Reads in one IBD record.  isExtendedFile should be true
	* iff this is an extended file. */
IBD IBDTable::readIBD(istream &is, Marker *m) {
	int n = _isExtendedFile ? 15 : 3;
	double p[n];  	
	
	for(int i = 0; i<n; i++)
		is >> p[i];    // FIXME: should check that there's only numbers
	skipPastChar(is, '\n');
	// FIXME: should check that there's nothing before that newline
	
	if (_isExtendedFile)
		// XXX this looks bizarre, but I wanted to emphasize that
		// the standard "extended" state names are 1-based
		return IBD(m, p[15-1], p[11-1], p[10-1], p[9-1]);
	else
		return IBD(m, p[0], p[1], p[2]);
}
