
#include <assert.h>

#include <algorithm>

#include "Pedigree.h"
#include "Utils.h"

Pedigree::Pedigree(istream &pedFile, istream &mapFile) {
	readMapFile(mapFile);
	readPedFile(pedFile);
}

void Pedigree::readMapFile(istream &mapFile) {
	int genotypeIndex = 0;
	double locInCm;
	
  string chr;
  while (mapFile >> chr) {
    Marker *m = new Marker();
    m->chrom = chr;
    mapFile >> m->id >> locInCm;
// cerr << m->chrom << " " << m->id << " " << locInCm << endl;
    
    // Convert the marker ID to lowercase.
		std::transform(m->id.begin(), m->id.end(), m->id.begin(), ::tolower);
		
    m->loc = (int) (locInCm * 1000000);
    m->genotypeIndex = genotypeIndex;
    
    // The markers are indexed two ways
    markersByLoc[m->chrom][m->loc] = m;
    markersByID[m->id] = m;
    
		genotypeIndex++;
		skipPastChar(mapFile, '\n');
	}

	numMarkers = genotypeIndex;
	cerr << "read marker file: numMarkers = " << numMarkers << endl;
}

void Pedigree::readPedFile(istream &pedFile) {

	// these are to store the parent IDs
	map<pair<string, string>, string> fatherID;
	map<pair<string, string>, string> motherID;
	string f,m,family;
  while (pedFile >> family) {
  	Person *p = new Person();
  	p->family = family;
		pedFile >> p->id >> f >> m >> p->sex;
// cerr << p->family << " " << p->id << " " << f << " " << m << " " << p->sex << " ";
		fatherID[pair<string,string>(p->family, p->id)] = f;
		motherID[pair<string,string>(p->family, p->id)] = m;

		// these will get filled in later
		p->father = NULL;
		p->mother = NULL;

		// read in genotypes
  	int allele1, allele2;
  	for(int i=0; i<numMarkers; i++) {
    	pedFile >> allele1;
    	skipToDigit(pedFile);   // skip to next digit, past whitespace or '/' (or anything else)
    	pedFile >> allele2;
    	p->genotype.push_back(Genotype(allele1, allele2));
// cerr << allele1 << " " << allele2 << "  ";
    }
  
  	// FIXME: check for duplicate people
  	peopleByID[ pair<string, string>(p->family, p->id) ] = p;
// cerr << endl;
	skipPastChar(pedFile, '\n');
  }

	// "link up" the parent pointers
	for(map<pair<string,string>, Person *>::iterator p=peopleByID.begin();
			p != peopleByID.end(); p++) {
		pair<string, string> id = p->first;   // the "family, ID" pair
		string family = id.first;
		
		if (fatherID[id] != "x" && fatherID[id] != "0") {
			Person * father =
				peopleByID[ pair<string,string>(family, fatherID[id]) ];
			// if this is NULL, exit with an error message
			if (father)
				p->second->father = father;
			else {
				cout << "couldn't find father for person " <<
					id.first << " " << id.second << endl;
				exit(1);
			}
		}
		
		if (motherID[id] != "x" && motherID[id] != "0") {
			Person * mother =
				peopleByID[ pair<string,string>(family, motherID[id]) ];
			// if this is NULL, exit with an error message
			if (mother)
				p->second->mother = mother;
			else {
				cout << "couldn't find mother for person " <<
					id.first << " " << id.second << endl;
				exit(1);
			}
		}
	}

}

void Pedigree::writePedToStream(ostream &os) {
	for(map<pair<string, string>, Person *>::iterator i =
			peopleByID.begin();
			i != peopleByID.end(); i++)
// 		cout << i->first.first << " " << i->first.second << endl;
		writePerson(i->second, os);
}

void Pedigree::writePerson(Person *p, ostream &os) {
	os << p->family << " " << p->id << " ";
	os << (p->father ? p->father->id : "x") << " ";
	os << (p->mother ? p->mother->id : "x") << " ";
	os << p->sex << " ";
	for(int i = 0; i < numMarkers; i++)
		os << p->genotype[i].a << "/" << p->genotype[i].b << " ";
	os << endl;
}

/** Writes out the genotypes in a pedigree in CSV format. */
void Pedigree::writePedigreeAsCSV(ostream &os) {
		for(map<pair<string, string>, Person *>::iterator it =
				peopleByID.begin();
				it != peopleByID.end(); it++) {
			Person *p = it->second;
			
			// for now, skip if the parents and grandparents aren't genotyped
//			if (!( (p->father) && (p->mother) && 
//					(p->father->father) && (p->father->mother) &&
//					(p->mother->father) && (p->mother->mother)))
//				continue;
			// for now, skip if parents of this person aren't genotyped
			if (!( (p->father) && (p->mother)))
				continue;	
			
			for(map<string, map<int, Marker *> >::iterator chrom
					= markersByLoc.begin();
					chrom != markersByLoc.end(); chrom++)
				for(map<int, Marker *>::iterator loc = chrom->second.begin();
						loc != chrom->second.end(); loc++) {
					Marker *m = loc->second;
					int i = loc->second->genotypeIndex;
					os << p->family << "," << p->id << "," << m->id << "," <<
						p->genotype[m->genotypeIndex].toSNP() << endl;
				}
		}
}

/* Write pedigree in the "compressed" format that we put in	the
	database: all the genotypes from everyone are all on one line. */
void Pedigree::writePedigreeCompressed(ostream &os) {

	// loop through the markers (in map order, though it doesn't
	// really matter)
	for(map<string, map<int, Marker *> >::iterator chrom
			= markersByLoc.begin();
			chrom != markersByLoc.end(); chrom++)
		for(map<int, Marker *>::iterator loc = chrom->second.begin();
				loc != chrom->second.end(); loc++) {
			Marker *m = loc->second;
			cout << m->id << ",";
			
			// loop through the people in the pedigree
			for(map<pair<string, string>, Person *>::iterator it =
					peopleByID.begin();
					it != peopleByID.end(); it++) {
				Person *p = it->second;
				// skip if it's not someone who was inferred
				// FIXME this is somewhat slow
				if (!( (p->father) && (p->mother) && 
						(p->father->father) && (p->father->mother) &&
						(p->mother->father) && (p->mother->mother)))
					continue;
				cout << p->genotype[ m->genotypeIndex ].toSNP();
			}
		cout << endl;
	}
}

