
#include <stdlib.h>

#include "SibGenotypeInferrer.h"

/** Constructor. */
SibGenotypeInferrer::SibGenotypeInferrer
(Pedigree &_ped, IBDTable &_ibd): ped(_ped), ibd(_ibd) {
		// Check that we've been given extended IBD states
		if (!ibd.isExtendedFile()) {
			cerr << "Error: to do sib inference, we need extended IBD states" << endl;
			exit(1);
		}
}

void SibGenotypeInferrer::inferGenotypes()  {

	// we loop through all people, and find all cases where both
	// parents are present.
	for(map<pair<string, string>, Person *>::iterator i =
			ped.peopleByID.begin();
			i != ped.peopleByID.end(); i++) {
			
		// are both parents present?	if not, skip to next person
		p = i->second;
		if (!( (father = p->father) && (mother = p->mother)))
			continue;

		// at this point, we know both parents are present, so we loop
		// through all sibs of p, and try to use that sib to infer
		// genotypes for p
		for(map<pair<string, string>, Person *>::iterator j =
				ped.peopleByID.begin();
				j != ped.peopleByID.end(); j++)
			if (j->second->father == father && j->second->mother == mother
					&& p != j->second)	{
				sib = j->second;
				inferGenotypesAtAllMarkers();
		}
	}
}

/** This will fill in one person's genotypes, using one sib.
	p, sib, father, and mother should all be set at this point. */
void SibGenotypeInferrer::inferGenotypesAtAllMarkers()  {
	for(map<string, map<int, Marker *> >::iterator chrom
			= ped.markersByLoc.begin(); chrom != ped.markersByLoc.end(); chrom++)
		for(map<int, Marker *>::iterator loc = chrom->second.begin();
				loc != chrom->second.end(); loc++)
			inferGenotype(*(loc->second));
}

void SibGenotypeInferrer::inferGenotype(Marker &m) {	
	int i = m.genotypeIndex;
	
	// if the genotype is already known, do nothing
	if (! p->genotype[i].isMissing() )
		return;

	short fatherAllele = inferAllele(m, p, sib, father, mother);
	short motherAllele = inferAllele(m, p, sib, mother, father);
	
  p->genotype[i] = Genotype(fatherAllele, motherAllele);
  
  // if the genotype is missing at this point, check for the case in
  // which both parents and the sib are heterozygous, and IBD is
  // known to be 0 or 2
  if (p->genotype[i].isMissing()) {
  	if (sib->genotype[i].isHeterozygous() &&
  			father->genotype[i].isHeterozygous() &&
  			mother->genotype[i].isHeterozygous()) {
  		IBD sharing = ibd.getIBD(p, sib, m.chrom, m.loc);
  		if (sharing.p0() + sharing.p2() == 1)
  			p->genotype[i] = sib->genotype[i];
  	}
  }
	// cout << "inferred genotype = " << p->genotype[i].toSNP() << endl;
}

/** Here, we're figuring out the allele transmitted by parent1;
	parent2 is just there to help resolve which allele was transmitted. */
short SibGenotypeInferrer::inferAllele
(Marker &m, Person *p, Person *sib, Person *parent1, Person *parent2) {
	int i = m.genotypeIndex;

	
//cout << "in SibGenotypeInferrer: inferring genotypes for " <<
//	p->family << " " << p->id << " at " << m.id << ", "<< m.chrom << ":" << m.loc
//	 	<< "  sib=" << sib->id
//		<< " p1=" << parent1->id << " p2=" << parent2->id << endl;

//	  cout << "genotypes: sib=" << sib->genotype[i].toSNP() << " " <<
//	"parent1=" << parent1->genotype[i].toSNP() << " " <<
//	"parent2=" << parent2->genotype[i].toSNP() << " " << endl;
	
	// If the parent genotype is missing, we can't infer the allele.
	// (For now, I'm leaving this check in, even though it's not always true.)
	if (parent1->genotype[i].isMissing())
		return 0;
		
	// If the parent is homozygous, we're done.
	if (parent1->genotype[i].isHomozygous()) {
//		cout << "parent is homozygous" << endl;
		return parent1->genotype[i].a;
	}
	
	// We'll need to know whether p and sib shared the parental allele or not		
	double sharingProb = ibd.getSharingOneSide(p, sib, m.chrom, m.loc,
		parent1->sex == 1);
	
	// for now, we only include cases in which IBD is known exactly
	if (!(sharingProb == 0 || sharingProb == 1))
		return 0;
	bool sharedSide = (sharingProb == 1.0);

//	cout << (parent1->sex==1?"pat.":"mat.") << " sharing=" << sharingProb << endl;

	// at this point, we'll need the child genotype
	if (sib->genotype[i].isMissing())
		return 0;

	// If the child is homozygous, and we know the sharing, then we're done.
	if (sib->genotype[i].isHomozygous()) {
//		cout << "sib is homozygous" << endl;
		short allele = sib->genotype[i].a;
		if (sharedSide)
			return allele;
		else
			return otherAllele(parent1->genotype[i], allele);
	}

	// At this point, we'll need the other parent's genotype.
	if (parent2->genotype[i].isMissing())
		return 0;

	// At this point, we know the child and parent1 are heterozygous.
	if (parent2->genotype[i].isHomozygous()) {
		// if the other parent is homozygous, we can tell what was transmitted
//		cout << "parent 2 is homozygous" << endl;
		short allele = parent2->genotype[i].a;
		if (sharedSide)
			return otherAllele(parent1->genotype[i], allele);
		else
			return allele;
	}
	else {
		// if all three people are heterozygous, we can't infer the allele here.
		return 0;
	}
}

/** Utility to return the other allele from a genotype. */
short SibGenotypeInferrer::otherAllele(Genotype g, short allele) {
	if (g.a == allele)
		return g.b;
	else
		return g.a;
}
