#!/usr/bin/perl -w
# Runs several tests on the genotype inference program.
# Input is the files in "sparse" and "dense".
# Requires the following in your $PATH:
# - "merlin" and "pedstats", available from Abecasis lab, at:
#   http://www.sph.umich.edu/csg/abecasis/Merlin/download/
#   http://www.sph.umich.edu/csg/abecasis/PedStats/download/
# - perl (tested using 5.8.7)

my $inferrer = "../../src/inferrer";

# First runs Merlin to estimate IBD; these IBD estimates will
# be stored in "merlin.ibd".
# Then, infers children's genotypes, given grandparents and parent genotypes.
system(<<END);
merlin --ibd --markerNames --singlepoint -d sparse/pedstats.dat -m sparse/pedstats.map -p sparse/pedstats.ped
$inferrer -i merlin.ibd -m dense/pedstats.map -p dense/pedstats.ped > inferred_3generation.ped
pedstats -d dense/pedstats.dat -p inferred_3generation.ped > inferred_3generation_pedstats.txt
END
	;

# Creates two-generation pedigree files, with just parents and sibs.
remove_founders("sparse/pedstats.ped", "sparse_2generation.ped", undef);
remove_founders("inferred_3generation.ped", "dense_2generation.ped", 1);

# First runs Merlin on the two-generation sparse pedigree to estimate
# extended IBD; these extended IBD estimates will be stored in "merlin.s15".
# Then, infers children's genotypes, given the genotypes for
# the parents and one sib.
system(<<END);
merlin --extended --markerNames --singlepoint -d sparse/pedstats.dat -m sparse/pedstats.map -p sparse_2generation.ped
$inferrer --infer sibs -i merlin.s15 -m dense/pedstats.map -p dense_2generation.ped > inferred_2generation.ped
pedstats -d dense/pedstats.dat -p inferred_2generation.ped > inferred_2generation_pedstats.txt
END
	;

# Removes the founders from a pedigree file, and optionally
# also removes most of the child genotypes.
# XXX this is a hack, that depends a lot on the specific
# identifiers (parents are 1 and 2, and 3 is a child.)
sub remove_founders {
	my($infile, $outfile, $remove_child_genotypes) = @_;

	open IN, "<$infile" || die;
	open OUT, ">$outfile" || die;
	
	while (<IN>) {
		chomp;
		next if /end/;
		/(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(.*)/;
		my $fam = $1;
		my $ind = $2;
		my $dad = $3;
		my $mom = $4;
		my $sex = $5;
		my $rest = $6;

		# remove grandparent genotypes, and references to grandparents
		if (($dad eq "0" && $mom eq "0") || ($dad eq "x" && $mom eq "x")) {
			next;
		}
		if ($ind eq "1" || $ind eq "2") {
			$dad = 0;
			$mom = 0;
		}

		# also, possibly remove genotypes of kids, _except_ for kid number 3
		if ($remove_child_genotypes) {
			if (($dad eq "1" && $mom eq "2") && (!($ind eq "3"))) {
				$rest =~ s/\d\s*\/\s*\d/0\/0/g;
			}
		}
		
		print OUT "$fam $ind $dad $mom $sex $rest\n";
	}

	close IN; close OUT;
}
